"""Acoupi-compatible Model that runs BirdNET."""

from acoupi import data
from acoupi.components import types
from audioclass.models import birdnet


class BirdNET(types.Model):
    """BirdNET Model to analyse the audio recording.

    This model uses the audioclass library to import the BirdNET model and
    identify and classify birds vocalisation in audio recordings.

    Attributes
    ----------
    name : str
        The name of the model, by default "BirdNET".
    """

    name = "BirdNET"

    def __init__(self, min_conf: float = 0.25, common_name: bool = False):
        """Initialize the model."""
        self.min_conf = min_conf
        self.common_name = common_name
        self._model = None

    @property
    def model(self) -> birdnet.BirdNET:
        """Get the BirdNET model."""
        if self._model is None:
            self._model = birdnet.BirdNET.load(
                confidence_threshold=self.min_conf,
                common_name=self.common_name,
            )
        return self._model

    def run(self, recording: data.Recording) -> data.ModelOutput:
        """Run the model on the recording.

        Parameters
        ----------
        recording : data.Recording
            The audio recording to process.

        Returns
        -------
        data.ModelOutput
            The model output containing the detections.
        """
        if not recording.path:
            return data.ModelOutput(
                name_model="BirdNET",
                recording=recording,
            )

        predictions = self.model.process_file(recording.path)
        high_freq = recording.samplerate / 2

        return data.ModelOutput(
            name_model="BirdNET",
            recording=recording,
            detections=[
                data.Detection(
                    detection_score=predicted_tag.score,
                    location=data.BoundingBox.from_coordinates(
                        prediction.clip.start_time,
                        0,
                        prediction.clip.end_time,
                        high_freq,
                    ),
                    tags=[
                        data.PredictedTag(
                            tag=data.Tag(
                                key=predicted_tag.tag.term.label,
                                value=predicted_tag.tag.value,
                            ),
                            confidence_score=predicted_tag.score,
                        )
                    ],
                )
                for prediction in predictions
                for predicted_tag in prediction.tags
            ],
        )



import datetime

from acoupi.components import MicrophoneConfig
from acoupi.programs.templates import MessagingConfig

from acoupi_birdnet.configuration import BirdNETConfig


def test_can_instantiate_default_configurations():
    config = BirdNETConfig(
        microphone=MicrophoneConfig(
            device_name="test_device",
        ),
        messaging=MessagingConfig(),
    )

    assert config.recording.schedule_start == datetime.time(hour=4)
    assert config.recording.schedule_end == datetime.time(hour=23)

    assert config.model.detection_threshold == 0.4



from acoupi_birdnet.configuration import BirdNETConfig
from acoupi_birdnet.model import BirdNET
from acoupi_birdnet.program import BirdNETProgram

__all__ = [
    "BirdNET",
    "BirdNETConfig",
    "BirdNETProgram",
]



"""BirdNET Program Configuration Options."""

import datetime
from typing import Optional

from acoupi.programs.templates import (
    AudioConfiguration,
    DetectionProgramConfiguration,
)
from pydantic import BaseModel, Field


class BirdNETAudioConfig(AudioConfiguration):
    """Audio Configuration schema."""

    schedule_start: datetime.time = Field(
        default=datetime.time(hour=4, minute=0, second=0),
    )
    """Start time for recording schedule."""

    schedule_end: datetime.time = Field(
        default=datetime.time(hour=23, minute=0, second=0),
    )
    """End time for recording schedule."""


class ModelConfig(BaseModel):
    """Model output configuration."""

    detection_threshold: float = 0.4
    """Detection threshold for filtering model outputs."""


class SaveRecordingFilter(BaseModel):
    """Saving Filters for audio recordings configuration."""

    starttime: datetime.time = datetime.time(hour=17, minute=0, second=0)
    """Start time of the interval for which to save recordings."""

    endtime: datetime.time = datetime.time(hour=21, minute=0, second=0)
    """End time of the interval for which to save recordings."""

    before_dawndusk_duration: int = 30
    """Optional duration in minutes before dawn/dusk to save recordings."""

    after_dawndusk_duration: int = 30
    """Optional duration in minutes after dawn/dusk to save recordings."""

    frequency_duration: int = 0
    """Optional duration in minutes to save recordings using the frequency filter."""

    frequency_interval: int = 0
    """Optional periodic interval in minutes to save recordings."""


class SaveRecordingManager(BaseModel):
    """Saving configuration for audio recordings.

    (path to storage, name of files, saving threshold).
    """

    true_dir: str = "birds"
    """Directory for saving recordings with confident detections."""

    false_dir: str = "no_birds"
    """Directory for saving recordings with uncertain detections."""

    timeformat: str = "%Y%m%d_%H%M%S"
    """Time format for naming the audio recording files."""

    saving_threshold: float = 0.2
    """Minimum threshold of detections from a recording to save it."""


class Summariser(BaseModel):
    """Summariser configuration."""

    interval: Optional[float] = 3600  # interval in seconds
    """Interval (in seconds) for summarising detections."""

    low_band_threshold: Optional[float] = 0.0
    """Optional low band threshold to summarise detections."""

    mid_band_threshold: Optional[float] = 0.0
    """Optional mid band threshold to summarise detections."""

    high_band_threshold: Optional[float] = 0.0
    """Optional high band threshold to summarise detections."""


class BirdNETConfig(DetectionProgramConfiguration):
    """BirdNET Program Configuration schema.

    This schema extends the _acoupi_ `DetectionProgramConfiguration` to
    include settings for the BirdNET program, such as custom audio recording,
    model setup, file management, messaging, and summarisation.
    """

    recording: AudioConfiguration = Field(
        default_factory=BirdNETAudioConfig,
    )
    """Audio recording configuration."""

    model: ModelConfig = Field(
        default_factory=ModelConfig,
    )
    """Model output configuration."""

    saving_filters: SaveRecordingFilter = Field(
        default_factory=SaveRecordingFilter,
    )
    """Recording Saving Filters configuration for audio recordings."""

    saving_managers: SaveRecordingManager = Field(
        default_factory=SaveRecordingManager,
    )
    """Recording Saving Managers configuration for audio recordings."""

    summariser_config: Optional[Summariser] = Field(
        default_factory=Summariser,
    )
    """Summariser configuration."""



"""Test Suite for Acoupi BirdNET Model."""

import datetime
from pathlib import Path

from acoupi import data

from acoupi_birdnet.model import BirdNET

TEST_RECORDING = Path(__file__).parent / "data" / "soundscape.wav"


def test_birdnet():
    recording = data.Recording(
        path=TEST_RECORDING,
        duration=2 * 60,
        samplerate=48_000,
        created_on=datetime.datetime.now(),
        deployment=data.Deployment(
            name="test",
        ),
    )

    model = BirdNET()
    detections = model.run(recording)

    assert isinstance(detections, data.ModelOutput)
    assert detections.name_model == "BirdNET"
    assert len(detections.detections) == 25



"""BirdNET Program."""

import datetime

import pytz
from acoupi import components, data, tasks
from acoupi.components import types
from acoupi.programs.templates import DetectionProgram

from acoupi_birdnet.configuration import BirdNETConfig
from acoupi_birdnet.model import BirdNET


class BirdNETProgram(DetectionProgram[BirdNETConfig]):
    """BirdNET Program Configuration."""

    config_schema = BirdNETConfig

    def setup(self, config: BirdNETConfig):
        """Set up the BirdNET Program.

        This method initialises the birdnet program, registers the
        recording, detection, management, messaging, and summariser tasks,
        and performs any necessary setup for the program to run.
        """
        # Setup all the elements from the DetectionProgram
        super().setup(config)

        if config.summariser_config and config.summariser_config.interval:
            summary_task = tasks.generate_summariser_task(
                summarisers=self.get_summarisers(config),
                message_store=self.message_store,
                logger=self.logger.getChild("summary"),
            )

            self.add_task(
                function=summary_task,
                schedule=datetime.timedelta(minutes=config.summariser_config.interval),
            )

    def configure_model(self, config):
        """Configure the BirdNET model.

        Returns
        -------
        BirdNET
            The BirdNET model instance.
        """
        return BirdNET()

    def get_summarisers(self, config) -> list[types.Summariser]:
        """Get the summarisers for the BirdNET Program.

        Parameters
        ----------
        config : BirdNET_ConfigSchema
            The configuration schema for the _acoupi_birdnet_ program defined in
            the configuration.py file and configured by a user via the CLI.

        Returns
        -------
        list[types.Summariser]
            A list of summarisers for the birdnet program. By default,
            the summariser will use the `summariser_config.interval` parameter for summarising
            the detections and calculating the minimum, maximum, and average
            confidence scores of the detections in each interval.
        """
        if not config.summariser_config:
            return []

        summarisers = []
        summariser_config = config.summariser_config

        if summariser_config.interval:
            summarisers.append(
                components.StatisticsDetectionsSummariser(
                    store=self.store,  # type: ignore
                    interval=summariser_config.interval,
                )
            )

        if (
            summariser_config.interval
            and summariser_config.low_band_threshold is not None
            and summariser_config.low_band_threshold != 0.0
            and summariser_config.mid_band_threshold is not None
            and summariser_config.mid_band_threshold != 0.0
            and summariser_config.high_band_threshold is not None
            and summariser_config.high_band_threshold != 0.0
        ):
            summarisers.append(
                components.ThresholdsDetectionsSummariser(
                    store=self.store,  # type: ignore
                    interval=summariser_config.interval,
                    low_band_threshold=summariser_config.low_band_threshold,
                    mid_band_threshold=summariser_config.mid_band_threshold,
                    high_band_threshold=summariser_config.high_band_threshold,
                )
            )

        return summarisers

    def get_file_managers(
        self, config: BirdNETConfig
    ) -> list[types.RecordingSavingManager]:
        """Get the file managers for the BirdNET Program.

        Parameters
        ----------
        config : BirdNET_ConfigSchema
            The configuration schema for the _acoupi_birdnet_ program defined in
            the configuration.py file and configured by a user via the CLI.

        Returns
        -------
        list[types.RecordingSavingManager]
            A list of file managers for the birdnet program.
        """
        return [
            components.SaveRecordingManager(
                dirpath=config.paths.recordings,
                dirpath_true=config.paths.recordings / config.saving_managers.true_dir,
                dirpath_false=config.paths.recordings
                / config.saving_managers.false_dir,
                timeformat=config.saving_managers.timeformat,
                detection_threshold=config.model.detection_threshold,
                saving_threshold=config.saving_managers.saving_threshold,
            )
        ]

    def get_message_factories(
        self, config: BirdNETConfig
    ) -> list[types.MessageBuilder]:
        """Get the message factories for the BirdNET Program.

        Parameters
        ----------
        config : BirdNET_ConfigSchema
            The configuration schema for the _acoupi_birdnet_ program defined in
            the configuration.py file and configured by a user via the CLI.

        Returns
        -------
        list[types.MessageBuilder]
            A list of message factories for the birdnet program. By default,
            the message factory will use the `detection_threshold` parameter for
            buildling messages.
        """
        return [
            components.DetectionThresholdMessageBuilder(
                detection_threshold=config.model.detection_threshold
            )
        ]

    def get_recording_filters(
        self, config: BirdNETConfig
    ) -> list[types.RecordingSavingFilter]:
        """Get the recording filters for the BirdNET Program.

        Parameters
        ----------
        config : BirdNET_ConfigSchema
            The configuration schema for the _acoupi_birdnet_ program defined in
            the configuration.py file and configured by a user via the CLI.

        Returns
        -------
        list[types.RecordingSavingFilter]
            A list of recording filters for the birdnet program. If no
            saving filters are defined, the method will not save any recordings.
        """
        if not config.saving_filters:
            # No saving filters defined
            return []

        saving_filters = []
        timezone = pytz.timezone(config.timezone)
        recording_saving = config.saving_filters

        # Main filter
        # Will only save recordings if the recording time is in the
        # interval defined by the start and end time.
        if (
            recording_saving is not None
            and recording_saving.starttime is not None
            and recording_saving.endtime is not None
        ):
            saving_filters.append(
                components.SaveIfInInterval(
                    interval=data.TimeInterval(
                        start=recording_saving.starttime,
                        end=recording_saving.endtime,
                    ),
                    timezone=timezone,
                )
            )

        # Additional filters
        if (
            recording_saving is not None
            and recording_saving.frequency_duration != 0
            and recording_saving.frequency_interval != 0
        ):
            # This filter will only save recordings at a frequency defined
            # by the duration and interval.
            saving_filters.append(
                components.FrequencySchedule(
                    duration=recording_saving.frequency_duration,
                    frequency=recording_saving.frequency_interval,
                )
            )

        if (
            recording_saving is not None
            and recording_saving.before_dawndusk_duration != 0
        ):
            # This filter will only save recordings if the recording time
            # is before dawn or dusk.
            saving_filters.append(
                components.Before_DawnDuskTimeInterval(
                    duration=recording_saving.before_dawndusk_duration,
                    timezone=timezone,
                )
            )

        if (
            recording_saving is not None
            and recording_saving.after_dawndusk_duration != 0
        ):
            # This filter will only save recordings if the recording time
            # is after dawn or dusk.
            saving_filters.append(
                components.After_DawnDuskTimeInterval(
                    duration=recording_saving.after_dawndusk_duration,
                    timezone=timezone,
                )
            )

        return saving_filters
